#include "common_header.h"
#include "OpenGLApp.h"
#include "Shaders.h"
#include "Texture.h"
#include "VertexBufferObject.h"
#include "static_geometry.h"

#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

/* One VBO, where all static data are stored now,
in this tutorial vertex is stored as 3 floats for
position and 2 floats for texture coordinate */

CVertexBufferObject vboSceneObjects;
UINT uiVAO; // And one VAO

CShader shVertex, shFragment;
CShaderProgram spMain;

CTexture tGold, tSnow;

/*-----------------------------------------------

Name:   InitScene

Params: lpParam - Pointer to anything you want.

Result: Initializes OpenGL features that will
        be used.

/*---------------------------------------------*/

void InitScene(LPVOID lpParam)
{
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	vboSceneObjects.CreateVBO();
	glGenVertexArrays(1, &uiVAO); // Create one VAO
	glBindVertexArray(uiVAO);

	vboSceneObjects.BindVBO();

	// Add cube to VBO

	FOR(i, 36)
	{
		vboSceneObjects.AddData(&vCubeVertices[i], sizeof(glm::vec3));
		vboSceneObjects.AddData(&vCubeTexCoords[i%6], sizeof(glm::vec2));
	}

	// Add pyramid to VBO

	FOR(i, 12)
	{
		vboSceneObjects.AddData(&vPyramidVertices[i], sizeof(glm::vec3));
		vboSceneObjects.AddData(&vPyramidTexCoords[i%3], sizeof(glm::vec2));
	}

	// Add ground to VBO

	FOR(i, 6)
	{
		vboSceneObjects.AddData(&vGround[i], sizeof(glm::vec3));
		vCubeTexCoords[i] *= 5.0f;
		vboSceneObjects.AddData(&vCubeTexCoords[i%6], sizeof(glm::vec2));
	}

	vboSceneObjects.UploadDataToGPU(GL_STATIC_DRAW);

	// Vertex positions start on zero index, and distance between two consecutive is sizeof whole
	// vertex data (position and tex. coord)
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(glm::vec3)+sizeof(glm::vec2), 0);
	// Texture coordinates start right after positon, thus on (sizeof(glm::vec3)) index,
	// and distance between two consecutive is sizeof whole vertex data (position and tex. coord)
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(glm::vec3)+sizeof(glm::vec2), (void*)sizeof(glm::vec3));

	// Load shaders and create shader program

	shVertex.LoadShader("data/shaders/shader.vert", GL_VERTEX_SHADER);
	shFragment.LoadShader("data/shaders/shader.frag", GL_FRAGMENT_SHADER);

	spMain.CreateProgram();
	spMain.AddShaderToProgram(&shVertex);
	spMain.AddShaderToProgram(&shFragment);

	spMain.LinkProgram();
	spMain.UseProgram();

	glEnable(GL_DEPTH_TEST);
	glClearDepth(1.0);

	// Finally, load our texture

	tGold.LoadTexture2D("data/textures/golddiag.jpg", true);
	tGold.SetFiltering(TEXTURE_FILTER_MAG_BILINEAR, TEXTURE_FILTER_MIN_BILINEAR_MIPMAP);

	tSnow.LoadTexture2D("data/textures/snow.jpg", true);
	tSnow.SetFiltering(TEXTURE_FILTER_MAG_BILINEAR, TEXTURE_FILTER_MIN_BILINEAR_MIPMAP);

	if (!CheckGLError("InitScene"))
		PRINT_INFO("Initialized scene with no errors");
}

/*-----------------------------------------------

Name:   RenderScene

Params: lpParam - Pointer to anything you want.

Result: Renders whole scene.

/*---------------------------------------------*/

// All angles stored in radians
float fRotationAngleCube = 0.0f, fRotationAnglePyramid = 0.0f;
float fCubeRotationSpeed = 0.0f, fPyramidRotationSpeed = 0.0f;

void DisplayTextureFiltersInfo()
{
	char buf[255];
	std::string sInfoMinification[] =
	{
		"Nearest",
		"Bilinear"
	};
	std::string sInfoMagnification[] =
	{
		"Nearest",
		"Bilinear",
		"Nearest on closest mipmap",
		"Bilinear on closest mipmap",
		"Trilinear"
	};
	sprintf(buf, "Mag. Filter: %s, Min. Filter: %s", sInfoMinification[tSnow.GetMagnificationFilter()].c_str(),
	                                                 sInfoMagnification[tSnow.GetMinificationFilter()-2].c_str());
	SDL_SetWindowTitle(appMain.window, buf);
}

void RenderScene(LPVOID lpParam)
{
	// Typecast lpParam to COpenGLControl pointer
	COpenGLControl* oglControl = (COpenGLControl*)lpParam;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	int iModelViewLoc = glGetUniformLocation(spMain.GetProgramID(), "modelViewMatrix");
	int iProjectionLoc = glGetUniformLocation(spMain.GetProgramID(), "projectionMatrix");
	glUniformMatrix4fv(iProjectionLoc, 1, GL_FALSE, glm::value_ptr(*oglControl->GetProjectionMatrix()));

	glm::mat4 mModelView = glm::lookAt(glm::vec3(0, 12, 27), glm::vec3(0, 0, 0), glm::vec3(0.0f, 1.0f, 0.0f));
	glm::mat4 mCurrent;

	glBindVertexArray(uiVAO);

	// Texture binding - we set GL_ACTIVE_TEXTURE0, and then we tell fragment shader,
	// that gSampler variable will fetch data from GL_ACTIVE_TEXTURE0

	int iSamplerLoc = glGetUniformLocation(spMain.GetProgramID(), "gSampler");
	glUniform1i(iSamplerLoc, 0);

	tGold.BindTexture(0);

	// Rendering of cube

	mCurrent = glm::translate(mModelView, glm::vec3(-8.0f, 0.0f, 0.0f));
	mCurrent = glm::scale(mCurrent, glm::vec3(10.0f, 10.0f, 10.0f));
	mCurrent = glm::rotate(mCurrent, fRotationAngleCube, glm::vec3(1.0f, 0.0f, 0.0f));
	glUniformMatrix4fv(iModelViewLoc, 1, GL_FALSE, glm::value_ptr(mCurrent));

	glDrawArrays(GL_TRIANGLES, 0, 36);

	// Rendering of pyramid

	mCurrent = glm::translate(mModelView, glm::vec3(8.0f, 0.0f, 0.0f));
	mCurrent = glm::scale(mCurrent, glm::vec3(10.0f, 10.0f, 10.0f));
	mCurrent = glm::rotate(mCurrent, fRotationAnglePyramid, glm::vec3(0.0f, 1.0f, 0.0f));
	glUniformMatrix4fv(iModelViewLoc, 1, GL_FALSE, glm::value_ptr(mCurrent));

	glDrawArrays(GL_TRIANGLES, 36, 12);

	// Render ground

	tSnow.BindTexture();

	glUniformMatrix4fv(iModelViewLoc, 1, GL_FALSE, glm::value_ptr(mModelView));
	glDrawArrays(GL_TRIANGLES, 48, 6);

	// A little interaction for user

	const unsigned char * state = SDL_GetKeyboardState(NULL);
	if (state[SDL_SCANCODE_UP])
		fCubeRotationSpeed -= appMain.sof(glm::radians(60.0f));
	if (state[SDL_SCANCODE_DOWN])
		fCubeRotationSpeed += appMain.sof(glm::radians(60.0f));
	if (state[SDL_SCANCODE_LEFT])
		fPyramidRotationSpeed -= appMain.sof(glm::radians(60.0f));
	if (state[SDL_SCANCODE_RIGHT])
		fPyramidRotationSpeed += appMain.sof(glm::radians(60.0f));

	fRotationAngleCube += appMain.sof(fCubeRotationSpeed);
	fRotationAnglePyramid += appMain.sof(fPyramidRotationSpeed);

	oglControl->SwapBuffersM();

	CheckGLError("RenderScene");
}

/*-----------------------------------------------

Name:   keyboard

Params: event - Pointer to SDL Event that triggered
        this function.

Result: none

/*---------------------------------------------*/

void keyboard(SDL_Event * event)
{
	// F1 and F2 change the texture filterings and set window text about that
	switch (event->key.keysym.sym) {
		case SDLK_F1:
			tGold.SetFiltering((tGold.GetMagnificationFilter()+1)%2, tGold.GetMinificationFilter());
			tSnow.SetFiltering((tSnow.GetMagnificationFilter()+1)%2, tSnow.GetMinificationFilter());
			DisplayTextureFiltersInfo();
			break;
		case SDLK_F2:
			int iNewMinFilter = tSnow.GetMinificationFilter() == TEXTURE_FILTER_MIN_TRILINEAR ? TEXTURE_FILTER_MIN_NEAREST : tSnow.GetMinificationFilter()+1;
			tSnow.SetFiltering(tSnow.GetMagnificationFilter(), iNewMinFilter);
			tGold.SetFiltering(tGold.GetMagnificationFilter(), iNewMinFilter);
			DisplayTextureFiltersInfo();
			break;
	}
}

/*-----------------------------------------------

Name:   ReleaseScene

Params: lpParam - Pointer to anything you want.

Result: Releases OpenGL scene.

/*---------------------------------------------*/

void ReleaseScene(LPVOID lpParam)
{
	spMain.DeleteProgram();

	shVertex.DeleteShader();
	shFragment.DeleteShader();

	vboSceneObjects.ReleaseVBO();
	glDeleteVertexArrays(1, &uiVAO);

	tGold.ReleaseTexture();
	tSnow.ReleaseTexture();
}
