#include "common_header.h"
#include "OpenGLControl.h"

#include <glm/gtc/matrix_transform.hpp>

/*-----------------------------------------------

Name:   InitGLEW

Params: none

Result: Creates fake window and OpenGL rendering
        context, within which GLEW is initialized.

/*---------------------------------------------*/

bool COpenGLControl::InitGLEW()
{
	if (bGlewInitialized)
		return true;

	bool bResult = true;
	if(!bGlewInitialized)
	{
		glewExperimental = true;
		if(glewInit() != GLEW_OK)
		{
			PRINT_ERROR("Couldn't initialize GLEW!");
			bResult = false;
		}
		glGetError(); // Discard GLEW errors
		bGlewInitialized = true;
	}

	return bResult;
}

/*-----------------------------------------------

Name:   InitOpenGL

Params: a_window - window to init OpenGL into
        a_initScene - pointer to init function
        a_renderScene - pointer to render function
        a_releaseScene - optional parameter of release function

Result: Initializes OpenGL rendering context
        of specified version. If succeeds, returns true.

/*---------------------------------------------*/

bool COpenGLControl::InitOpenGL(SDL_Window * a_window, void (*a_InitScene)(void *), void (*a_RenderScene)(void *), void(*a_ReleaseScene)(void *), void * lpParam)
{
	window = a_window;
	context = SDL_GL_CreateContext(window);
	if (!context) {
		PRINT_ERROR("Error creating OpenGL Context");
		return false;
	}
	SDL_GL_SetSwapInterval(1);

	PRINT_INFO("OpenGL Version: " << glGetString(GL_VERSION));

	if (!InitGLEW())
		return false;

	RenderScene = a_RenderScene;
	InitScene = a_InitScene;
	ReleaseScene = a_ReleaseScene;

	if (InitScene != NULL)
		InitScene(lpParam);

	return true;
}

/*-----------------------------------------------

Name:   InitSDL2

Params: a_iMajorVersion - Major version of OpenGL
        a_iMinorVersion - Minor version of OpenGL

Result: Sets various hints to let SDL know what
        options we want in our OpenGL window

/*---------------------------------------------*/

bool COpenGLControl::InitSDL2(int iMajorVersion, int iMinorVersion)
{
	SDL_Init(SDL_INIT_VIDEO);

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, iMajorVersion);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, iMinorVersion);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);

	return true;
}

/*-----------------------------------------------

Name:   ResizeOpenGLViewportFull

Params: none

Result: Resizes viewport to full window with perspective
        projection.

/*---------------------------------------------*/

void COpenGLControl::ResizeOpenGLViewportFull()
{
	int w, h;
	if (window) {
		SDL_GetWindowSize(window, &w, &h);
		glViewport(0, 0, w, h);
		iViewportWidth = w;
		iViewportHeight = h;
	}
}

/*-----------------------------------------------

Name:   SetProjection3D

Params: fFOV - field of view angle
        fAspectRatio - aspect ration of width/height
        fNear, fFar - distance of near and far clipping plane

Result: Calculates projection matrix.

/*---------------------------------------------*/

void COpenGLControl::SetProjection3D(float fFOV, float fAspectRatio, float fNear, float fFar)
{
	mProjection = glm::perspective(fFOV, fAspectRatio, fNear, fFar);
}

/*-----------------------------------------------

Name:   SetOrtho2D

Params: width - width of window
        height - height of window

Result: Calculates ortho 2D projection matrix.

/*---------------------------------------------*/

void COpenGLControl::SetOrtho2D(int width, int height)
{
	mOrtho = glm::ortho(0.0f, float(width), 0.0f, float(height));
}

/*-----------------------------------------------

Name:   GetProjectionMatrix()

Params: none

Result: Retrieves pointer to projection matrix

/*---------------------------------------------*/

glm::mat4* COpenGLControl::GetProjectionMatrix()
{
	return &mProjection;
}

/*-----------------------------------------------

Name:   GetOrthoMatrix()

Params: none

Result: Retrieves pointer to ortho matrix.

/*---------------------------------------------*/

glm::mat4* COpenGLControl::GetOrthoMatrix()
{
	return &mOrtho;
}

/*-----------------------------------------------

Name:   SwapBuffersM

Params: none

Result: Swaps back and front buffer.

/*---------------------------------------------*/

void COpenGLControl::SwapBuffersM()
{
	SDL_GL_SwapWindow(window);
}

/*-----------------------------------------------

Name:   Render

Params: lpParam - pointer to whatever you want

Result: Calls previously set render function.

/*---------------------------------------------*/

void COpenGLControl::Render(void * lpParam)
{
	unsigned int tCurrent = SDL_GetTicks();
	if(tCurrent - tLastSecond > 1000)
	{
		tLastSecond += 1000;
		iFPSCount = iCurrentFPS;
		iCurrentFPS = 0;
	}

	if(RenderScene)
		RenderScene(lpParam);

	iCurrentFPS++;
}

/*-----------------------------------------------

Name:   ReleaseOpenGLControl

Params: lpParam - pointer to whatever you want

Result: Calls previously set release function
        and deletes rendering context.

/*---------------------------------------------*/

void COpenGLControl::ReleaseOpenGLControl(void * lpParam)
{
	if(ReleaseScene)
		ReleaseScene(lpParam);

	SDL_GL_DeleteContext(context);
	SDL_DestroyWindow(window);
	SDL_Quit();
	window = NULL;
}

/*-----------------------------------------------

Name:   SetVerticalSynchronization

Params: bEnabled - whether to enable V-Sync

Result: Guess what it does :)

/*---------------------------------------------*/

bool COpenGLControl::SetVerticalSynchronization(int interval)
{
	return (!SDL_GL_SetSwapInterval(interval));
}

/*-----------------------------------------------

Name:   Getters

Params: none

Result: ... They get something :D

/*---------------------------------------------*/

int COpenGLControl::GetFPS()
{
	return iFPSCount;
}

int COpenGLControl::GetViewportWidth()
{
	return iViewportWidth;
}

int COpenGLControl::GetViewportHeight()
{
	return iViewportHeight;
}
